// Block Decomposition Method implementation in Pascal ver. 1.0
// All rights reserved Hector Zenil and Fernando Soler-Toscano
// Algorithmic Nature Group (www.algorithmicnature.org)
// 
// The software on this web site, whether in the form of source code or executable, is provided As Is.
// BDM is Free Software distributed under the GNU General Public License.
// This means you are free to distribute and modify citing the source and making your own version available under the same 
// GPL license. Be sure to read the license. http://www.gnu.org/licenses/gpl.html 

program HashTableLookUp;

{$mode objfpc}{$H+}
//{$DEFINE DEBUG}

uses
  {$IFDEF UNIX}{$IFDEF UseCThreads}
  cthreads,
  {$ENDIF}{$ENDIF}
  Classes, contnrs, sysutils, strutils, math, dateutils, DefaultHashTable
  { you can add units after this };
Type
  TValueCount = Record
                    sKey:String;
                    dValue: Double;
                    wCount: Word;
  	        end;
  PFPStringHashTable = ^TFPStringHashTable;
  TAValueCount = Array of TValueCount;
Const
  HASH_TABLE_COMMANDLINE_OPTION = '-t';
  VALUE_FILE_COMMANDLINE_OPTION = '-v';
  HELP_COMMANDLINE_OPTION = '-h';

  CSVFILE_SEPARATOR = ',';
  KEY_SEPARATOR = '-';
  VALUE_SEPARATOR = KEY_SEPARATOR;
  DECIMALPOINT_SEPARATOR = '.';

// Procedures and functions
function PopulateHashTable(sFileName: String; pHashTable: PFPStringHashTable; var wKeySubStringLenght, wKeySubStringCount: Word):Boolean;
var
  fNewHashTableFile: Text;  // The file to be read for creating the table
  oKeyValuePair: TStringList;  // List of strings to contain the key and value pair for each file line
  lwInputLines: Longword;   // Counter to keep track of the lines in the file and adjust hash table size accordingly
  sReadLine: String;        // String to contain each line read from the file
begin
  PopulateHashTable := False;
  try
      // Open file for reading
     try
        Assign(fNewHashTableFile, sFileName);
        Reset(fNewHashTableFile);
     Except
        WriteLn('ERROR: failed to open ' + sFileName + ' for reading');
        Halt;
     end;

     // Variables initializatron
     oKeyValuePair := TStringList.Create;
     oKeyValuePair.Delimiter := CSVFILE_SEPARATOR;
     lwInputLines := 0;
     sReadLine := '';
     wKeySubStringLenght := 0;
     wKeySubStringCount := 0;

     // Read lines from file, split them into key-value pairs and add them to the table
     while Not(EOF(fNewHashTableFile)) do begin
        ReadLn(fNewHashTableFile, sReadLine);
        if not(AnsiCompareStr(sReadLine, '') = 0) then begin
           oKeyValuePair.DelimitedText := sReadLine;
           lwInputLines := lwInputLines + 1;
           pHashTable^.Add(oKeyValuePair[0], oKeyValuePair[1]);
        end;
     end; // end WHILE

     // Get key length from the last line read
     oKeyValuePair.Delimiter := KEY_SEPARATOR;
     oKeyValuePair.DelimitedText := oKeyValuePair.Strings[0];
     wKeySubStringLenght := strLen(Pchar(oKeyValuePair.Strings[0]));
     wKeySubStringCount := oKeyValuePair.Count;

     Close(fNewHashTableFile);
     PopulateHashTable := True;
  Except
    WriteLn('ERROR: error populating hash table using data from file ' + sFileName + slineBreak + 'Last line read is No.' + IntToStr(lwInputLines) +
                    ': ' + sReadLine);
    Halt;
  end;
end; // end PopulateHashTable
procedure CaluclateFinalValue(oHashTable: TFPStringHashTable; sValue: String; wKeySubStringLenght, wKeySubstringCount:Word);
var i, i1, i2 : Word;
    iFromSubString, iToSubString, iEndTableElement: Int64;
    iDesignatedSubStrings, iSubstringPortions: LongInt;
    sSubStrings: TStringList;  // List of strings to contain the substrings of the value string
    aTempTable: TAValueCount;  // Auxiliary table used for storing unique values from the main hash table and their count
    sTemp: String;
    dFinalValue: Double;
begin
  Try
     // Create auxiliary, temp table
      try
         SetLength(aTempTable, 1);
         aTempTable[0].dValue := 0;
         aTempTable[0].wCount := 0;
         aTempTable[0].sKey   := '';

         {$IFDEF DEBUG}
	    WriteLn('Temp real values table succesfully created');
	 {$ENDIF}
      Except
        WriteLn('ERROR: Unable to create or populate temporary table');
        Halt;
      end;
     sSubStrings := TStringList.Create;
     sSubStrings.Delimiter := VALUE_SEPARATOR;

     sSubStrings.DelimitedText := sValue;
     iDesignatedSubStrings := (sSubStrings.Count div wKeySubstringCount)-1;
     iSubstringPortions := (Length(sSubStrings[0]) div wKeySubStringLenght)-1;
       for i1 := 0 to iDesignatedSubStrings do begin
           iFromSubString := Int64(i1)*wKeySubstringCount;
           iToSubString := iFromSubString + wKeySubstringCount - 1;
           for i2 := 0 to iSubstringPortions do begin
                   sValue := '';
	           for i := iFromSubString to iToSubString do begin
	              sValue := sValue + MidStr(sSubStrings[i], (wKeySubStringLenght*i2)+1, wKeySubStringLenght) + KEY_SEPARATOR;
	           end;  // end FOR i

	           sValue := LeftStr(sValue, Length(sValue) - 1);

	           sTemp := oHashTable.Items[sValue];
	           if AnsiCompareStr(sValue, '') = 0 then begin
	              WriteLn('missing value');
	              Halt;
	           end
	           else begin
	              {$IFDEF DEBUG}
			 WriteLn('Key: ' + sValue + ' -> ' + sTemp);
		      {$ENDIF}

                      iEndTableElement := High(aTempTable);
                      dFinalValue := StrToFloat(sTemp);

                      // Look for the real value in the table
                      for i := 0 to iEndTableElement do
                         if AnsiCompareStr(aTempTable[i].sKey, sValue) = 0 then begin
                             // Increase counter for this real value
                             aTempTable[i].wCount := aTempTable[i].wCount + 1;
                             Break;
                         end;
	              if i >= iEndTableElement then  begin// Real value not found in temp table
	                  SetLength(aTempTable, iEndTableElement + 2);
                          aTempTable[iEndTableElement].wCount :=  1;
                          aTempTable[iEndTableElement].sKey := sValue;
                          aTempTable[iEndTableElement].dValue := dFinalValue;
                      end;
	           end;
           end; // end i2
       end;  // end FOR i1


     iEndTableElement := Int64(High(aTempTable)) - 1;
    {$IFDEF DEBUG}
        WriteLn;
        WriteLn('Real Values Temp Table:');
        for i := 0 to iEndTableElement do begin
	    Writeln('Key: ' + aTempTable[i].sKey + '; Value: '+ FloatToStr(aTempTable[i].dValue) + '; Count: ' + IntToStr(aTempTable[i].wCount));
        end;

        WriteLn;
        WriteLn('Final Calculated Value:');
    {$ENDIF}

// Calculate actual value

     dFinalValue := 0;
     // Iterate through real values table
     for i := 0 to iEndTableElement do begin
         //dFinalValue := dFinalValue + aTempTable[i].dValue + ((aTempTable[i].dValue)*log2(aTempTable[i].wCount));
         dFinalValue := dFinalValue + aTempTable[i].dValue + (log2(aTempTable[i].wCount));
     end;
     WriteLn(FloatToStr(dFinalValue));
  Except
    WriteLn('ERROR: error trying to calculate final real value');
    Halt;
  end;
end; // end CaluclateFinalValue
Procedure PrintHelp();
begin
     WriteLn('Usage: ' + ParamStr(0) + '[value] [' + VALUE_FILE_COMMANDLINE_OPTION + '] filename [' + HASH_TABLE_COMMANDLINE_OPTION + ' filename]');
     WriteLn('"[value]" specifies string value to be used for composing hash keys. Use double quotes around the value, just in case! It''s mandatory to provide this or [' + VALUE_FILE_COMMANDLINE_OPTION + '] parameter');
     WriteLn('"' + VALUE_FILE_COMMANDLINE_OPTION + ' filename" specifies file with input values. It''s mandatory to provide this or [value] parameter');
     WriteLn('"' + HASH_TABLE_COMMANDLINE_OPTION + ' filename" specifies file with hash key-value pairs. If omitted, default built-in hash table will be used');
     WriteLn('"' + HELP_COMMANDLINE_OPTION + '" prints this message');
end;

var
  oHashTable: TFPStringHashTable;  // Hash table

  sValue: String;                    // String to contain each value line
  sValuesFileName: String;	    // the name of the file containing the values
  fValuesFile: Text;
  sHashTableFileName: String;	    // the name of the file to be read for creating new table on the fly (if provided)
  wKeySubStringLenght, wKeySubStringCount: Word;       // Legnght and count of the hashtable key subtrings
  i: LongInt;

begin
  Try
      if ParamCount < 1 then begin
         PrintHelp;
         Exit;
      end;
      // Initialize variables
      sValuesFileName := '';
      sHashTableFileName := '';
      DefaultFormatSettings.DecimalSeparator := DECIMALPOINT_SEPARATOR;
      sValue := '';

      // Read parameters and populate corresponding variables
      i := 1;
      while i <= ParamCount do begin
        case ParamStr(i) of
    	     HASH_TABLE_COMMANDLINE_OPTION: begin
                                               i := i + 1;
                                               if FileExists(ParamStr(i)) then begin
                                                  sHashTableFileName := ParamStr(i);
                                              end
                                              else begin
                                                  WriteLn('ERROR: ' + ParamStr(i + 1) + ' file not found');
                                                  Exit;
                                              end;

                                            end;  //end -t option
             VALUE_FILE_COMMANDLINE_OPTION: begin
                                                   i := i + 1;
                                                   if FileExists(ParamStr(i)) then begin
                                                      sValuesFileName := ParamStr(i);
	                                           end
	                                           else begin
	                                              WriteLn('ERROR: ' + ParamStr(i + 1) + ' file not found');
	                                              Exit;
	                                           end;
         				      end;  // end -v option
             HELP_COMMANDLINE_OPTION:  begin
                                            PrintHelp;
                                            Exit;
                                       end;
             else
               sValue := ParamStr(i);

        end; // end CASE
        i := i + 1;
      end;  // end WHILE

      // Check if we have at least the mandatory value argument
      if ((AnsiCompareStr(sValuesFileName, '') = 0) And
         (AnsiCompareStr(sValue, '') = 0)) then begin
         WriteLn('ERROR: no value to search in the hashtable found. Make sure you provide [value] or [' + VALUE_FILE_COMMANDLINE_OPTION + '] argument');
         PrintHelp;
         Exit;
      end;

      {$IFDEF DEBUG}
         WriteLn('Command line parameters:');
         WriteLn('HashTable File Name:' + #9 + sHashTableFileName);
         WriteLn('Values File Name:' + #9 + sValuesFileName);
         WriteLn;
      {$ENDIF}
      // Create hash table
      try
         oHashTable := TFPStringHashTable.Create;
         wKeySubStringLenght := 0;
         wKeySubStringCount := 0;

         if AnsiCompareStr(sHashTableFileName, '') = 0 then begin
             // Set file to default
             {$IFDEF DEBUG}
	      WriteLn('Trying to populate hastable using built-in values');
	     {$ENDIF}
             for i := 0 to DEFAULT_HASHTABLE_ROWS_LASTINDEX do
                oHashTable.Add(DEFAULT_HASHTABLE[i][0],DEFAULT_HASHTABLE[i][1]);

             wKeySubStringLenght := DEFAULT_HASHTABLE_KEYSUBSTRINGLENGHT;
             wKeySubStringCount := DEFAULT_HASHTABLE_KEYSUBSTRINGCOUNT;
         end
         else begin
              {$IFDEF DEBUG}
	      WriteLn('Trying to populate hashtable using file ' + sHashTableFileName);
	      {$ENDIF}
             if Not(PopulateHashTable(sHashTableFileName, @oHashTable, wKeySubStringLenght, wKeySubStringCount)) Or
                (wKeySubStringLenght = 0 ) Or
                (wKeySubStringCount = 0 ) then begin
                    WriteLn('ERROR: error populating hash table using data from file ' + sHashTableFileName);
                    Exit;
             end;
         end;

         {$IFDEF DEBUG}
	    WriteLn('Hash table succesfully created');
	 {$ENDIF}

      Except
        WriteLn('ERROR: Unable to create or populate hashtable');
        Exit;
      end;

      if (AnsiCompareStr(sValuesFileName, '') = 0) then begin
         CaluclateFinalValue(oHashTable, sValue, wKeySubStringLenght, wKeySubstringCount);
      end
      else begin
          // Open values file for reading
          try
            Assign(fValuesFile, sValuesFileName);
            Reset(fValuesFile);

            {$IFDEF DEBUG}
	        WriteLn('Succesfully opened for reading values file ' + sValuesFileName);
	    {$ENDIF}
          Except
            WriteLn('ERROR: Failed to open ' + sValuesFileName + ' for reading');
            Exit;
          end;

          // Variables initializatron
          sValue := '';

          {$IFDEF DEBUG}
             WriteLn;
             WriteLn('Begin reading values from input file:');
          {$ENDIF}

          // Read lines from file, split them into key-value pairs and add them to the table
          while Not(EOF(fValuesFile)) do begin
            ReadLn(fValuesFile, sValue);
            {$IFDEF DEBUG}
                WriteLn;
	        WriteLn('Line: ' + sValue);
            {$ENDIF}
            CaluclateFinalValue(oHashTable, sValue, wKeySubStringLenght, wKeySubstringCount);
          end; // end WHILE

          Close(fValuesFile);

          {$IFDEF DEBUG}
          WriteLn('Finished reading input file');
          {$ENDIF}
      end; // end IF value file name is empty
      {$IFDEF DEBUG}
         WriteLn;
         WriteLn('Program end!');
      {$ENDIF}
  Except
    WriteLn('ERROR: Unknown error occured');
  end;

end.

